@@ -24,18 +24,19 @@
   */
 #include "rsync.h"
 
-static int total_written = 0;
-static int total_read = 0;
+static int64 total_written;
+static int64 total_read;
 
 extern int verbose;
 extern int sparse_files;
+extern int io_timeout;
 
-int write_total(void)
+int64 write_total(void)
 {
   return total_written;
 }
 
-int read_total(void)
+int64 read_total(void)
 {
   return total_read;
 }
@@ -49,10 +50,10 @@ void setup_nonblocking(int f_in,int f_out)
 }
 
 
-static char *read_buffer = NULL;
-static char *read_buffer_p = NULL;
-static int read_buffer_len = 0;
-static int read_buffer_size = 0;
+static char *read_buffer;
+static char *read_buffer_p;
+static int read_buffer_len;
+static int read_buffer_size;
 
 
 /* This function was added to overcome a deadlock problem when using
@@ -95,39 +96,70 @@ static void read_check(int f)
   }
 }
 
+static time_t last_io;
+
+
+static void check_timeout(void)
+{
+	time_t t;
+	
+	if (!io_timeout) return;
+
+	if (!last_io) {
+		last_io = time(NULL);
+		return;
+	}
+
+	t = time(NULL);
+
+	if (last_io && io_timeout && (t-last_io)>io_timeout) {
+		fprintf(FERROR,"read timeout after %d second - exiting\n", 
+			(int)(t-last_io));
+		exit_cleanup(1);
+	}
+}
 
 static int readfd(int fd,char *buffer,int N)
 {
-  int  ret;
-  int total=0;  
+	int  ret;
+	int total=0;  
+	struct timeval tv;
+	
+	if (read_buffer_len < N)
+		read_check(buffer_f_in);
+	
+	while (total < N) {
+		if (read_buffer_len > 0 && buffer_f_in == fd) {
+			ret = MIN(read_buffer_len,N-total);
+			memcpy(buffer+total,read_buffer_p,ret);
+			read_buffer_p += ret;
+			read_buffer_len -= ret;
+			continue;
+		} 
 
-  if (read_buffer_len < N)
-	  read_check(buffer_f_in);
- 
-  while (total < N)
-    {
-      if (read_buffer_len > 0 && buffer_f_in == fd) {
-	ret = MIN(read_buffer_len,N-total);
-	memcpy(buffer+total,read_buffer_p,ret);
-	read_buffer_p += ret;
-	read_buffer_len -= ret;
-      } else {
-	while ((ret = read(fd,buffer + total,N - total)) == -1) {
-	  fd_set fds;
+		while ((ret = read(fd,buffer + total,N-total)) == -1) {
+			fd_set fds;
 
-	  if (errno != EAGAIN && errno != EWOULDBLOCK)
-	    return -1;
-	  FD_ZERO(&fds);
-	  FD_SET(fd, &fds);
-	  select(fd+1, &fds, NULL, NULL, NULL);
+			if (errno != EAGAIN && errno != EWOULDBLOCK)
+				return -1;
+			FD_ZERO(&fds);
+			FD_SET(fd, &fds);
+			tv.tv_sec = io_timeout;
+			tv.tv_usec = 0;
+
+			if (select(fd+1, &fds, NULL, NULL, &tv) != 1) {
+				check_timeout();
+			}
+		}
+
+		if (ret <= 0)
+			return total;
+		total += ret;
 	}
-      }
 
-      if (ret <= 0)
+	if (io_timeout)
+		last_io = time(NULL);
 	return total;
-      total += ret;
-    }
-  return total;
 }
 
 
@@ -145,6 +177,34 @@ int read_int(int f)
   return IVAL(b,0);
 }
 
+int64 read_longint(int f)
+{
+	extern int remote_version;
+	int64 ret;
+	char b[8];
+	ret = read_int(f);
+
+	if (ret != -1) return ret;
+
+#ifndef HAVE_LONGLONG
+	fprintf(FERROR,"Integer overflow - attempted 64 bit offset\n");
+	exit_cleanup(1);
+#else
+	if (remote_version >= 16) {
+		if ((ret=readfd(f,b,8)) != 8) {
+			if (verbose > 1) 
+				fprintf(FERROR,"(%d) Error reading %d bytes : %s\n",
+					getpid(),8,ret==-1?strerror(errno):"EOF");
+			exit_cleanup(1);
+		}
+		total_read += 8;
+		ret = IVAL(b,0) | (((int64)IVAL(b,4))<<32);
+	}
+#endif
+
+	return ret;
+}
+
 void read_buf(int f,char *buf,int len)
 {
   int ret;
@@ -165,8 +225,8 @@ unsigned char read_byte(int f)
 }
 
 
-static char last_byte=0;
-static int last_sparse = 0;
+static char last_byte;
+static int last_sparse;
 
 int sparse_end(int f)
 {
@@ -234,6 +294,7 @@ static int writefd(int fd,char *buf,int len)
 	    /* hmmm, we got a write select on the fd and then failed to write.
 	       Why doesn't that mean that the fd is dead? It doesn't on some
 	       systems it seems (eg. IRIX) */
+	    u_sleep(1000);
 #if 0
 	    fprintf(FERROR,"write exception\n");
 	    exit_cleanup(1);
@@ -266,7 +327,10 @@ static int writefd(int fd,char *buf,int len)
 	      exit_cleanup(1);
       }
 
-      if (count == 0) continue;
+      if (count == 0) {
+	      check_timeout();
+	      continue;
+      }
       
       if (FD_ISSET(fd, &w_fds)) {
 	      got_select = 1;
@@ -276,6 +340,9 @@ static int writefd(int fd,char *buf,int len)
     }
   }
 
+  if (io_timeout)
+	  last_io = time(NULL);
+
   return total;
 }
 
@@ -294,6 +361,29 @@ void write_int(int f,int x)
   total_written += 4;
 }
 
+void write_longint(int f, int64 x)
+{
+	extern int remote_version;
+	char b[8];
+	int ret;
+
+	if (remote_version < 16 || x <= 0x7FFFFFFF) {
+		write_int(f, (int)x);
+		return;
+	}
+
+	write_int(f, -1);
+	SIVAL(b,0,(x&0xFFFFFFFF));
+	SIVAL(b,4,((x>>32)&0xFFFFFFFF));
+
+	if ((ret=writefd(f,b,8)) != 8) {
+		fprintf(FERROR,"write_longint failed : %s\n",
+			ret==-1?strerror(errno):"EOF");
+		exit_cleanup(1);
+	}
+	total_written += 8;
+}
+
 void write_buf(int f,char *buf,int len)
 {
   int ret;
