@@ -22,7 +22,7 @@
 int verbose = 0;
 int always_checksum = 0;
 time_t starttime;
-off_t total_size = 0;
+int64 total_size = 0;
 int block_size=BLOCK_SIZE;
 
 char *backup_suffix = BACKUP_SUFFIX;
@@ -54,26 +54,29 @@ int am_root=0;
 int orig_umask=0;
 int relative_paths=0;
 int numeric_ids = 0;
+int force_delete = 0;
+int io_timeout = 0;
+int io_error = 0;
 
 extern int csum_length;
 
 int am_server = 0;
-static int sender = 0;
+static int sender;
 int recurse = 0;
 
 static void usage(FILE *f);
 
 static void report(int f)
 {
-  int in,out,tsize;
+  int64 in,out,tsize;
   time_t t = time(NULL);
   
   if (!verbose) return;
 
   if (am_server && sender) {
-    write_int(f,read_total());
-    write_int(f,write_total());
-    write_int(f,total_size);
+    write_longint(f,read_total());
+    write_longint(f,write_total());
+    write_longint(f,total_size);
     write_flush(f);
     return;
   }
@@ -81,17 +84,24 @@ static void report(int f)
   if (sender) {
     in = read_total();
     out = write_total();
-    tsize = (int)total_size;
+    tsize = total_size;
   } else {
-    in = read_int(f);
-    out = read_int(f);
-    tsize = read_int(f);
+    in = read_longint(f);
+    out = read_longint(f);
+    tsize = read_longint(f);
   }
 
-  printf("wrote %d bytes  read %d bytes  %g bytes/sec\n",
-	 out,in,(in+out)/(0.5 + (t-starttime)));        
-  printf("total size is %d  speedup is %g\n",
-	 tsize,(1.0*tsize)/(in+out));
+#if HAVE_LONGLONG
+  printf("wrote %lld bytes  read %lld bytes  %g bytes/sec\n",
+	 (long long)out,(long long)in,(in+out)/(0.5 + (t-starttime)));
+  printf("total size is %lld  speedup is %g\n",
+	 (long long)tsize,(1.0*tsize)/(in+out));
+#else
+  printf("wrote %ld bytes  read %ld bytes  %g bytes/sec\n",
+	 (long)out,(long)in,(in+out)/(0.5 + (t-starttime)));
+  printf("total size is %ld  speedup is %g\n",
+	 (long)tsize,(1.0*tsize)/(in+out));
+#endif
 }
 
 
@@ -100,6 +110,7 @@ static void server_options(char **args,int *argc)
   int ac = *argc;
   static char argstr[50];
   static char bsize[30];
+  static char iotime[30];
   int i, x;
 
   args[ac++] = "--server";
@@ -160,6 +171,11 @@ static void server_options(char **args,int *argc)
     args[ac++] = bsize;
   }    
 
+  if (io_timeout) {
+    sprintf(iotime,"--timeout=%d",io_timeout);
+    args[ac++] = iotime;
+  }    
+
   if (strcmp(backup_suffix, BACKUP_SUFFIX)) {
 	  args[ac++] = "--suffix";
 	  args[ac++] = backup_suffix;
@@ -168,6 +184,9 @@ static void server_options(char **args,int *argc)
   if (delete_mode)
     args[ac++] = "--delete";
 
+  if (force_delete)
+    args[ac++] = "--force";
+
   if (numeric_ids)
     args[ac++] = "--numeric-ids";
 
@@ -185,7 +204,7 @@ static int do_cmd(char *cmd,char *machine,char *user,char *path,int *f_in,int *f
 {
   char *args[100];
   int i,argc=0, ret;
-  char *tok,*p,*dir=NULL;
+  char *tok,*dir=NULL;
 
   if (!local_server) {
     if (!cmd)
@@ -220,31 +239,18 @@ static int do_cmd(char *cmd,char *machine,char *user,char *path,int *f_in,int *f
 
   server_options(args,&argc);
 
-  if (path && *path) {
-    dir = strdup(path);
-    p = strrchr(dir,'/');
-    if (p && !relative_paths) {
-      *p = 0;
-      if (!dir[0])
-	args[argc++] = "/";
-      else
-	args[argc++] = dir;
-      p++;
-    } else {
-      args[argc++] = ".";
-      p = dir;
-    }
-    if (p[0])
-      args[argc++] = path;
-  }
+  args[argc++] = ".";
+
+  if (path && *path) 
+	  args[argc++] = path;
 
   args[argc] = NULL;
 
   if (verbose > 3) {
-    fprintf(FERROR,"cmd=");
+    fprintf(FINFO,"cmd=");
     for (i=0;i<argc;i++)
-      fprintf(FERROR,"%s ",args[i]);
-    fprintf(FERROR,"\n");
+      fprintf(FINFO,"%s ",args[i]);
+    fprintf(FINFO,"\n");
   }
 
   ret = piped_child(args,f_in,f_out);
@@ -306,31 +312,31 @@ static char *get_local_name(struct file_list *flist,char *name)
 void do_server_sender(int argc,char *argv[])
 {
   int i;
-  char *dir = argv[0];
   struct file_list *flist;
+  char *dir = argv[0];
 
   if (verbose > 2)
-    fprintf(FERROR,"server_sender starting pid=%d\n",(int)getpid());
+    fprintf(FINFO,"server_sender starting pid=%d\n",(int)getpid());
   
   if (!relative_paths && chdir(dir) != 0) {
-    fprintf(FERROR,"chdir %s: %s (3)\n",dir,strerror(errno));
-    exit_cleanup(1);
+	  fprintf(FERROR,"chdir %s: %s (3)\n",dir,strerror(errno));
+	  exit_cleanup(1);
   }
   argc--;
   argv++;
   
   if (strcmp(dir,".")) {
-    int l = strlen(dir);
-    if (strcmp(dir,"/") == 0) 
-      l = 0;
-    for (i=0;i<argc;i++)
-      argv[i] += l+1;
+	  int l = strlen(dir);
+	  if (strcmp(dir,"/") == 0) 
+		  l = 0;
+	  for (i=0;i<argc;i++)
+		  argv[i] += l+1;
   }
 
   if (argc == 0 && recurse) {
-    argc=1;
-    argv--;
-    argv[0] = ".";
+	  argc=1;
+	  argv--;
+	  argv[0] = ".";
   }
     
 
@@ -356,10 +362,10 @@ static int do_recv(int f_in,int f_out,struct file_list *flist,char *local_name)
   }
   
 
-  if ((pid=fork()) == 0) {
+  if ((pid=do_fork()) == 0) {
     recv_files(f_in,flist,local_name,recv_pipe[1]);
     if (verbose > 2)
-      fprintf(FERROR,"receiver read %d\n",read_total());
+      fprintf(FINFO,"receiver read %ld\n",(long)read_total());
     exit_cleanup(0);
   }
 
@@ -374,21 +380,22 @@ static int do_recv(int f_in,int f_out,struct file_list *flist,char *local_name)
 void do_server_recv(int argc,char *argv[])
 {
   int status;
-  char *dir = NULL;
   struct file_list *flist;
   char *local_name=NULL;
+  char *dir = NULL;
   
   if (verbose > 2)
-    fprintf(FERROR,"server_recv(%d) starting pid=%d\n",argc,(int)getpid());
+    fprintf(FINFO,"server_recv(%d) starting pid=%d\n",argc,(int)getpid());
 
   if (argc > 0) {
-    dir = argv[0];
-    argc--;
-    argv++;
-    if (chdir(dir) != 0) {
-      fprintf(FERROR,"chdir %s : %s (4)\n",dir,strerror(errno));
-      exit_cleanup(1);
-    }    
+	  dir = argv[0];
+	  argc--;
+	  argv++;
+	  if (chdir(dir) != 0) {
+		  fprintf(FERROR,"chdir %s : %s (4)\n",
+			  dir,strerror(errno));
+		  exit_cleanup(1);
+	  }    
   }
 
   if (delete_mode)
@@ -401,11 +408,11 @@ void do_server_recv(int argc,char *argv[])
   }
 
   if (argc > 0) {    
-    if (strcmp(dir,".")) {
-      argv[0] += strlen(dir);
-      if (argv[0][0] == '/') argv[0]++;
-    }
-    local_name = get_local_name(flist,argv[0]);
+	  if (strcmp(dir,".")) {
+		  argv[0] += strlen(dir);
+		  if (argv[0][0] == '/') argv[0]++;
+	  }
+	  local_name = get_local_name(flist,argv[0]);
   }
 
   status = do_recv(STDIN_FILENO,STDOUT_FILENO,flist,local_name);
@@ -444,7 +451,9 @@ static void usage(FILE *f)
   fprintf(f,"    --rsync-path PATH    specify path to rsync on the remote machine\n");
   fprintf(f,"-C, --cvs-exclude        auto ignore files in the same way CVS does\n");
   fprintf(f,"    --delete             delete files that don't exist on the sending side\n");
+  fprintf(f,"    --force              force deletion of directories even if not empty\n");
   fprintf(f,"    --numeric-ids        don't map uid/gid values by user/group name\n");
+  fprintf(f,"    --timeout TIME       set IO timeout in seconds\n");
   fprintf(f,"-I, --ignore-times       don't exclude files that match length and time\n");
   fprintf(f,"-T  --temp-dir DIR       create temporary files in directory DIR\n");
   fprintf(f,"-z, --compress           compress file data\n");
@@ -459,7 +468,8 @@ static void usage(FILE *f)
 }
 
 enum {OPT_VERSION,OPT_SUFFIX,OPT_SENDER,OPT_SERVER,OPT_EXCLUDE,
-      OPT_EXCLUDE_FROM,OPT_DELETE,OPT_NUMERIC_IDS,OPT_RSYNC_PATH};
+      OPT_EXCLUDE_FROM,OPT_DELETE,OPT_NUMERIC_IDS,OPT_RSYNC_PATH,
+      OPT_FORCE,OPT_TIMEOUT};
 
 static char *short_options = "oblLWHpguDCtcahvrRIxnSe:B:T:z";
 
@@ -468,6 +478,7 @@ static struct option long_options[] = {
   {"server",      0,     0,    OPT_SERVER},
   {"sender",      0,     0,    OPT_SENDER},
   {"delete",      0,     0,    OPT_DELETE},
+  {"force",       0,     0,    OPT_FORCE},
   {"numeric-ids", 0,     0,    OPT_NUMERIC_IDS},
   {"exclude",     1,     0,    OPT_EXCLUDE},
   {"exclude-from",1,     0,    OPT_EXCLUDE_FROM},
@@ -497,6 +508,7 @@ static struct option long_options[] = {
   {"rsh",         1,     0,    'e'},
   {"suffix",      1,     0,    OPT_SUFFIX},
   {"block-size",  1,     0,    'B'},
+  {"timeout",     1,     0,    OPT_TIMEOUT},
   {"temp-dir",    1,     0,    'T'},
   {"compress",	  0,	 0,    'z'},
   {0,0,0,0}};
@@ -519,11 +531,6 @@ int main(int argc,char *argv[])
     struct file_list *flist;
     char *local_name = NULL;
 
-#ifdef SETPGRP_VOID
-    setpgrp();
-#else
-    setpgrp(0,0);
-#endif
     signal(SIGUSR1, sigusr1_handler);
 
     starttime = time(NULL);
@@ -531,7 +538,7 @@ int main(int argc,char *argv[])
 
     /* we set a 0 umask so that correct file permissions can be
        carried across */
-    orig_umask = umask(0);
+    orig_umask = (int)umask(0);
 
     while ((opt = getopt_long(argc, argv, 
 			      short_options, long_options, &option_index)) 
@@ -563,6 +570,10 @@ int main(int argc,char *argv[])
 	  delete_mode = 1;
 	  break;
 
+	case OPT_FORCE:
+	  force_delete = 1;
+	  break;
+
 	case OPT_NUMERIC_IDS:
 	  numeric_ids = 1;
 	  break;
@@ -690,6 +701,10 @@ int main(int argc,char *argv[])
 	  block_size = atoi(optarg);
 	  break;
 
+	case OPT_TIMEOUT:
+	  io_timeout = atoi(optarg);
+	  break;
+
 	case 'T':
 		tmpdir = optarg;
 		break;
@@ -781,7 +796,7 @@ int main(int argc,char *argv[])
     }
 
     if (verbose > 3) {
-      fprintf(FERROR,"cmd=%s machine=%s user=%s path=%s\n",
+      fprintf(FINFO,"cmd=%s machine=%s user=%s path=%s\n",
 	      shell_cmd?shell_cmd:"",
 	      shell_machine?shell_machine:"",
 	      shell_user?shell_user:"",
@@ -803,7 +818,7 @@ int main(int argc,char *argv[])
 #endif
 
     if (verbose > 3) 
-      fprintf(FERROR,"parent=%d child=%d sender=%d recurse=%d\n",
+      fprintf(FINFO,"parent=%d child=%d sender=%d recurse=%d\n",
 	      (int)getpid(),pid,sender,recurse);
 
     if (sender) {
@@ -813,10 +828,10 @@ int main(int argc,char *argv[])
 	send_exclude_list(f_out);
       flist = send_file_list(f_out,argc,argv);
       if (verbose > 3) 
-	fprintf(FERROR,"file list sent\n");
+	fprintf(FINFO,"file list sent\n");
       send_files(flist,f_out,f_in);
       if (verbose > 3)
-	fprintf(FERROR,"waiting on %d\n",pid);
+	fprintf(FINFO,"waiting on %d\n",pid);
       waitpid(pid, &status, 0);
       report(-1);
       exit_cleanup(status);
@@ -826,7 +841,7 @@ int main(int argc,char *argv[])
 
     flist = recv_file_list(f_in);
     if (!flist || flist->count == 0) {
-      fprintf(FERROR,"nothing to do\n");
+      fprintf(FINFO,"nothing to do\n");
       exit_cleanup(0);
     }
 
